/***********************************************************************

  Copyright (C) 2006  Adrien Brochier <a.brochier@free.fr>

  This file is part of SADE (Sivia Applied to Dielectric spectroscopy Experiment).

  SADE is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published
  by the Free Software Foundation; either version 2 of the License,
  or (at your option) any later version.

  SADE is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston,
  MA  02111-1307  USA

************************************************************************/

/**
  \file main.cpp
  \author Adrien Brochier <a.brochier@free.fr>
  \date 07-06
 */

#include <boost/thread/thread.hpp>
#include <boost/thread/condition.hpp>
#include <iostream>
#include <fstream>
#include <stdio.h>
#include <stack>
#include <deque>
#include <vector>
#include "pave.h"
#include "debye.h"
#include <time.h>
#include <string>

using namespace std;

	/*global variable declaration===============================================================*/
	
	pave param, 			/**<initial parameter*/
		contraintesR,		/**<real values of the data*/
		contraintesI,		/**<imaginary part of the data*/
		intersect;			/**<main result of the algorithme*/
	float precision;		/**<precision of the resultat*/
	float incertitude;		/**<accuracy of the data, in %. if it's <0, the accuracy is read in the data file*/
	int nbData;			/**<number of data in the data file*/
	int relaxMax;			/**<maximum number of relaxation*/
	int relaxMin;			/**<miniimum number of relaxation*/
	int relax;				/**<current number of relaxation*/
	
	interval a(0,2000);		/**<initial interval for \f$\epsilon_{\infty}\f$*/
	interval b(0.000000001,200000000);	/**<initial interval for the \f$\epsilon_i\f$*/
	interval c(0.0000000001,100000000);	/**<initial interval for the \f$\tau_i\f$*/
	interval sigma(0,1);					/**<initial interval for the conductivity*/
	vector <REEL> x;		/**<frequence read in the data file*/
	//int compteur=0;
	time_t debut,end, now1,now2; 
	stack<pave> pile;		/**<main stack*/
	stack<pave> stackResult;
	stack<short int> profondeur;/**<TODO : virer ca*/
	short int cur_prof=0;
	fstream file;
	fstream outFile;
	REEL f, yr, yi, er, ei, cont1, cont2;
	boost::mutex verrou;		/**<mutex for acces to the main stack*/
	boost::mutex chrono;		/**<mutex for the acces to the time variable*/
	boost::mutex mutCurrent;	/**<mutex for the access to current result (ie : intersect)*/
	int atWork = 0; 		/**<count the number of thread who are really working. if the stack is empty and atwork==0, there is nothing else to do*/
	boost::condition stackNotFree;	/**<thread condition : make the thread waiting while the pile is empty but the job isn't finished*/
	int threadCount=0;

/**
*\brief fonction passed to the threads.

This fonction implement the fitting algorithm, based on SIVIA. Basically, it takes a pave of parameter from the main stack, and test if it's feasible or not.
if yes, it store it. if no, it reject it. if it's indeterminate, it cut the current pave in 2 part and store each part ine the stack
*/

void processusStep1()
{
	int i,
		in,			/**< \brief une chouette variable
                      *
                      */
		overlap,		/**<flag. overlap==1 if the current pave intersect the solution*/
		threadID;		/**<unique integer who identify the thread. DEBUG*/
	interval temp,temp2;
	pave  current,		/**<current tested pave*/
		out1,
		out2,
		old,
		eps;
	boost::mutex::scoped_lock lock(verrou,boost::defer_lock); 
	boost::mutex::scoped_lock chronLock(chrono,boost::defer_lock);
	boost::mutex::scoped_lock mutLock(mutCurrent,boost::defer_lock);
	lock.lock();
	threadID=threadCount++;
	cout << "je suis le Thread " << threadID << " et je travaille..." << endl;
	lock.unlock();
	
	while ((!pile.empty()) || (atWork>0))
		{	
			lock.lock();
			while(pile.empty() && atWork > 0) stackNotFree.wait(lock);		 /*wait while the stack is empty*/
			if(pile.empty() && atWork ==0) return; 
			atWork++; 								/* we start a job, so we increment atWork*/
			current=pile.top();	pile.pop();				/*take the top of the stack*/
			lock.unlock();
			
			if(!(current < intersect))						/*if (current<intersect), we already know that current is "good"*/
			{
				in=1;
				overlap=1;
			
				/*Contractor=================================================*/
				do{
					old=current;		/*save the old value*/
					for(i=0;i<nbData;i++) 
					{
						
						current=contract(x[i],current,contraintesR[i],contraintesI[i],relax);
						
						if(isEmpty(current)) break;
					}
				}while( (mieux(current,old)) && ( !(isEmpty(current)) ) );	/*we do this while the new pave is better than the old one*/
				
				/*inclusion test ===============================================*/
				for(i=0;i<nbData;i++)
				{	
					eps=eval(x[i], current,relax);			/*compute the value corresponding to the current pave*/
					
					if(!((eps[0] < contraintesR[i])&&(eps[1] < contraintesI[i])))/*-> if current is not feasible*/
					{
						in=0;					/*flag in set to FALSE*/
						temp=eps[0]&contraintesR[i];
						temp2=eps[1]&contraintesI[i];
						if((temp.isEmpty())||(temp2.isEmpty()))	/* -> true => current unfeasible / false => current indeterminate*/
						{
							overlap=0;							
						}
					}
					
					if(!overlap) break;
				}
				
				if(in)			/*if in, current is feasible*/
				{
					mutLock.lock();
					stackResult.push(current);	
					mutLock.unlock();
				}else{
					if(overlap)
					{
						if(lengthTau(current)<=precision)	/*indeterminate but small enough to be accepted*/
						{
							mutLock.lock();
							stackResult.push(current);
							mutLock.unlock();
							
						}else{	/*bissect current, and push the 2 part in the stack.*/
							bissectionTau(current, out1, out2);
							lock.lock();
							pile.push(out1);
							pile.push(out2);		
							lock.unlock();
							stackNotFree.notify_all(); 	/*say to thread who are waiting that the stack isn't empty that they can work*/
						}
					}
				}
				
			}
			/*information output, every X seconds*/
			chronLock.lock();
			time(&now2);
			if(difftime(now2,now1) >=300)
			{
				cout << "jusqu'ici tout va bien" << endl;
				cout << "Here is the Thread "<<threadID << ", we are working since " << difftime(now2,debut) << " Sec" << endl;
				cout << "there are " << pile.size() << " items in the stack" << endl;
				affiche(intersect);
				cout << endl;
				time(&now1);
			}
			chronLock.unlock();
			lock.lock();
			atWork--;		/*end of current work/ we will wait for a new one*/
			lock.unlock();
			
			if(pile.empty() && atWork < 1) stackNotFree.notify_all();	/*to avoid a deadlock at the end*/			
		}	
	cout << "je suis le Thread " << threadID << " et j'ai fini !!" << endl;	
}
void processusStep2()
{
	int i,
		in,			/**< \brief une chouette variable
                      *
                      */
		overlap,		/**<flag. overlap==1 if the current pave intersect the solution*/
		threadID;		/**<unique integer who identify the thread. DEBUG*/
	interval temp,temp2;
	pave  current,		/**<current tested pave*/
		out1,
		out2,
		old,
		eps;
	boost::mutex::scoped_lock lock(verrou,boost::defer_lock); 
	boost::mutex::scoped_lock chronLock(chrono,boost::defer_lock);
	boost::mutex::scoped_lock mutLock(mutCurrent,boost::defer_lock);
	lock.lock();
	threadID=threadCount++;
	cout << "je suis le Thread " << threadID << " et je travaille..." << endl;
	lock.unlock();
	
	while ((!pile.empty()) || (atWork>0))
		{	
			lock.lock();
			while(pile.empty() && atWork > 0) stackNotFree.wait(lock);		 /*wait while the stack is empty*/
			if(pile.empty() && atWork ==0) return; 
			atWork++; 								/* we start a job, so we increment atWork*/
			current=pile.top();	pile.pop();				/*take the top of the stack*/
			lock.unlock();
			
			if(!(current < intersect))						/*if (current<intersect), we already know that current is "good"*/
			{
				in=1;
				overlap=1;
			
				/*Contractor=================================================*/
				do{
					old=current;		/*save the old value*/
					for(i=0;i<nbData;i++) 
					{
						
						current=contract(x[i],current,contraintesR[i],contraintesI[i],relax);
						
						if(isEmpty(current)) break;
					}
				}while( (mieux(current,old)) && ( !(isEmpty(current)) ) );	/*we do this while the new pave is better than the old one*/
				
				/*inclusion test ===============================================*/
				for(i=0;i<nbData;i++)
				{	
					eps=eval(x[i], current,relax);			/*compute the value corresponding to the current pave*/
					
					if(!((eps[0] < contraintesR[i])&&(eps[1] < contraintesI[i])))/*-> if current is not feasible*/
					{
						in=0;					/*flag in set to FALSE*/
						temp=eps[0]&contraintesR[i];
						temp2=eps[1]&contraintesI[i];
						if((temp.isEmpty())||(temp2.isEmpty()))	/* -> true => current unfeasible / false => current indeterminate*/
						{
							overlap=0;							
						}
					}
					
					if(!overlap) break;
				}
				
				if(in)			/*if in, current is feasible*/
				{
					mutLock.lock();
					intersect=intersect|current; 	/*we make the union of current and the main result*/
					mutLock.unlock();
				}else{
					if(overlap)
					{
						if(length(current)<=precision)	/*indeterminate but small enough to be accepted*/
						{
							mutLock.lock();
							intersect=intersect|current;
							mutLock.unlock();
							
						}else{	/*bissect current, and push the 2 part in the stack.*/
							bissection(current, out1, out2);
							lock.lock();
							pile.push(out1);
							pile.push(out2);		
							lock.unlock();
							stackNotFree.notify_all(); 	/*say to thread who are waiting that the stack isn't empty that they can work*/
						}
					}
				}
				
			}
			/*information output, every X seconds*/
			chronLock.lock();
			time(&now2);
			if(difftime(now2,now1) >=300)
			{
				cout << "jusqu'ici tout va bien" << endl;
				cout << "Here is the Thread "<<threadID << ", we are moulining since " << difftime(now2,debut) << " Sec" << endl;
				cout << "there are " << pile.size() << " items in the stack" << endl;
				affiche(intersect);
				cout << endl;
				time(&now1);
			}
			chronLock.unlock();
			lock.lock();
			atWork--;		/*end of current work/ we will wait for a new one*/
			lock.unlock();
			
			if(pile.empty() && atWork < 1) stackNotFree.notify_all();	/*to avoid a deadlock at the end*/			
		}	
	cout << "je suis le Thread " << threadID << " et j'ai fini !!" << endl;	
}
int main(int argc, char **argv)
{
	/*initialization==================================================*/
	
	/*Banner*/
	cout << endl;
	cout << "******************************************************************************" << endl;
	cout << "** Sivia Applied to Dielectric spectroscopy Experimentation, version 0.4.0a **" << endl;
	cout << "******************************************************************************" << endl;
	cout << "(Multi-thread version)" << endl;
	cout << endl;
	if(argc < 5) 
	{
		cout << "usage : sade nbRelaxationMin nbRelaxationMax errorMode precision dataFile" << endl;
		return 1;
	}
	
	/*set variable====================================================*/
	precision=atof(argv[4]);
	incertitude=atof(argv[3]);
	relaxMax=atoi(argv[2]);
	relaxMin=atoi(argv[1]);
	file.open(argv[5]);
	int i;
	pave eps;
	
	/*loading data====================================================*/
	
	if(!file.is_open())
	{
		cout << "File opening error : " << argv[5] << endl;
		return 1;
	}
	if(incertitude<=0)		/*the accuracy is in the data file*/
	{
		while((file >> f) && (file >> yr) && (file >> yi)&& (file >> er)&& (file >> ei))
		{
			x.push_back(f);
			contraintesR.push_back(interval(yr-1*er,yr+1*er));
			contraintesI.push_back(interval(yi-1*ei,yi+1*ei));//*/
		}
	}else{
		while((file >> f) && (file >> yr) && (file >> yi))
		{
			x.push_back(f);
			contraintesR.push_back(interval(yr*(1-incertitude/100),yr*(1+incertitude/100)));
			contraintesI.push_back(interval(yi*(1-incertitude/100),yi*(1+incertitude/100)));//*/
		}
	}
	file.close();
	nbData=x.size();
	cout << "\033[1;31mExperimental datas loaded...\033[0m" << endl<<endl;
	/*Main processus====================================================*/
	time(&now1);
	for(relax=relaxMin;relax<=relaxMax;relax++)
	{
		time(&debut);
		/*setting initial values*/
		param=pave();
		pile=stack<pave>();
		param.push_back(a);
		for(i=0;i<relax;i++)
		{
			param.push_back(c);
			param.push_back(b);
		}
		param.push_back(sigma);
		pile.push(param);
		threadCount=0;
		boost::thread_group threads;
		for(i=0;i<8;i++) threads.create_thread(&processusStep1); 		/*create and start the threads*/
			threads.join_all();
		pile=stackResult;
		threadCount=0;
		precision=5*precision; 						//TEST !!!!!!!!!!!!!!!!!!!
		for(i=0;i<8;i++) threads.create_thread(&processusStep2); 		/*create and start the threads*/
			threads.join_all();
		
		/*output=======================================================*/
		
		time(&end);
		cout << "Elapsed time : " << difftime(end,debut) <<" seconds" <<  endl;
		
		if(! isEmpty(intersect))
		{
			affiche( intersect ); cout << endl;		/*display result*/
			
			/*store the result in a file*/
			string fileName(argv[5]);
			fileName=fileName+".result";
			outFile.open(fileName.c_str(), ios::out|ios::trunc);
			outFile << "result for " << argv[5] << " : " << endl;
			outFile << "Computing time : " << difftime(end,debut) << " Sec" << endl;
			outFile << "Minimal box size : " << argv[4] << endl;
			outFile << "With a supposed error of : " << argv[3] << endl;
			outFile << "Eps_infty = " << intersect[0] << endl;
			for (i=1;i<2*relax+1;i+=2)
			{
				outFile <<endl<< "Tau_" << 1+i/2 << " = " << intersect[i] << endl;
				outFile << "deltaEps_" << 1+i/2 << " = " << intersect[i+1] << endl;
			}
			outFile << endl << "sigma" << " = " << intersect[2*relax+1] << endl;
			outFile.close();
		/*BUILDING AN OCTAVE / MATLAB FILE (debug)*/
			fileName=string(argv[5]);
			fileName=fileName+".octave";
			outFile.open (fileName.c_str(),ios::trunc|ios::out);
			for(i=0;i<nbData;i++)
			{
				eps=eval(x[i], intersect,relax);
				outFile << x[i] <<" " << eps[0].getFirst() << " " <<eps[1].getFirst() <<" " << eps[0].getLast() << " " <<eps[1].getLast() << endl ;
			}
			outFile.close();
			return 0;
				
		}else{	/*no result found*/
			cout << "Sorry, I was not able to find any feasible parameters with only " << relax << " relaxation...." << endl<<endl;
		}
	}
	return 0;
}
