/***********************************************************************

  Copyright (C) 2006  Adrien Brochier <a.brochier@free.fr>

  This file is part of SADE (Sivia Applied to Dielectric spectroscopy Experiment).

  SADE is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published
  by the Free Software Foundation; either version 2 of the License,
  or (at your option) any later version.

  SADE is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston,
  MA  02111-1307  USA

************************************************************************/

/**
  \file pave.cpp
  \author Adrien Brochier <a.brochier@free.fr>
  \date 07-06
 */

#include "pave.h"

/*constructor*/

interval::interval()
{
	first=0;
	last=0;
}
interval::interval(REEL f, REEL l)
{
	
	first=f;	//DEBUG -error;
	last=l;	//DEBUG +error;
}


/*ensemblist operations*/

interval interval::operator&(interval i)		//intersection
{
	if((*this) < i) return (*this);
	if(i < (*this)) return i;
	interval r;
	r.first=max(first, i.first);
	r.last=min(last, i.last);
	if(r.isEmpty()) return interval(1,-1);
	return r;
}
interval interval::operator|(interval i)		// convex union
{
	interval r;
	r.first=min(i.first, first);
	r.last=max(i.last, last);
	return r;
}


/*elementary operation*/
interval interval::operator+(interval i)
{
	return interval(first+i.first,last+i.last);
}
interval interval::operator+(REEL a)
{
	return interval(first+a-error,last+a+error);
}
interval interval::operator*(interval i)
{
	REEL a,b,c,d,f,l;
	a=first*i.first;
	b=first*i.last;
	c=last*i.first;
	d=last*i.last;
	
	f=min(min(a,b),min(c,d));
	l=max(max(a,b),max(c,d));
	return interval(f,l);
}
interval interval::operator*(REEL a)
{
	if(a>=0)
	{
		return interval(a*first, a*last);
	}else{
		return interval(a*last, a*first);
	}
}
interval interval::operator-(interval i)
{
	return interval(first-i.last,last-i.first);
}
interval interval::operator/(interval i)
{
	return (*this)*(1/i);
}
interval interval::operator^(REEL a)
{
	if(((int)a)==a && ((int)a)%2==0)
	{
		if(first*last > 0)
		{
			return interval(pow(first,a),pow(last,a));
		}else{
			return interval(0,max(pow(first,a),pow(last,a)));
		}
	}else{
		return interval(pow(first,a),pow(last,a));
	}
}
interval interval::operator^(interval i)
{
	interval a1=(*this)^i.first;
	interval a2=(*this)^i.last;
	return interval(min(a1.first,a2.first),max(a1.last,a2.last));
}

void interval::operator=(interval i)
{
	first=i.first;
	last=i.last;
}
int interval::operator<(interval i)
{
	return ( first >= i.first ) && ( last <= i.last ) ;
}
interval interval::operator%(REEL a)
{
	if((first >= 0) && (first <= a))
	{
		return (*this);
	}else{
		int k = (long) floorl(first/a);
		REEL var=k*a;
		return interval(first-var,last-var);
	}
}

//utils
ostream& operator<<(ostream& o, interval i)
{
	o << "[" << i.getFirst() << "," << i.getLast() << "]";
}

int interval::isEmpty(){return first >last;}
REEL interval::getFirst()
{
	return first;
}
REEL interval::getLast()
{
	return last;
}
REEL interval::length()
{
	if(isEmpty()) return -1;
	return last-first;
}

interval operator/(REEL a,interval& i)
{
	if( i.isEmpty() ) return i;
	if(i.first * i.last > 0 ) return interval(a/i.last,a/i.first);
	if(i.first==0) return interval(0,oo);
	if(i.last==0) return interval(-oo,0);
	
	return interval(-oo,+oo);
	
};
interval operator/(interval i, REEL a)
{
	return interval(i.first/a,i.last/a);
};
interval operator-(REEL a, interval i)
{
	return interval(a-i.first,a-i.last);
};





